# -*- coding: utf-8 -*-
"""
Tools for locating data or subarrays inside other arrays.
"""
import numpy as np


# FIXME: We need the str/repr formatting used in Numpy < 1.14.
try:
    np.set_printoptions(legacy="1.13")
except TypeError:
    pass


def find_vals(m, v):
    """
    Find all occurrences of all values in `v` in `m`

    Parameters
    ----------
    m : array_like
        Array to be searched.
    v : array_like
        Array of values to find in `m`.

    Returns
    -------
    ndarray
        True/False vector with True indicating position of any value
        in `v` within `m`. Will be all False if no values in `v` are
        in `m`.

    Notes
    -----
    `m` is flattened to 1d before searching (using column-major
    ordering 'F').  The values in `pv` correspond to::

          [  0      r  ...
             1    r+1
           ...    ...
           r-1   2r-1  ... r*c-1 ]  where m is r x c

    Examples
    --------
    >>> import numpy as np
    >>> from pyyeti import locate
    >>> m = np.array([[10, 20], [30, 20]])
    >>> locate.find_vals(m, 20)
    array([False, False,  True,  True], dtype=bool)
    >>> locate.find_vals(m, 30)
    array([False,  True, False, False], dtype=bool)
    >>> locate.find_vals(m, 100)
    array([False, False, False, False], dtype=bool)
    """
    m = np.atleast_1d(m)
    v = np.atleast_1d(v)
    m = m.ravel(order="F")
    v = v.ravel()
    pv = np.zeros(len(m), dtype=bool)
    for i in v:
        pv |= m == i
    return pv


def find_duplicates(v, tol=0.0):
    """
    Find duplicate values in a vector (or within a tolerance).

    Parameters
    ----------
    v : 1d array_like
        Vector to find duplicates in.
    tol : scalar; optional
        Tolerance for checking for duplicates. Values are considered
        duplicates if the absolute value of the difference is <=
        `tol`.

    Returns
    -------
    dups : 1d ndarray
        Bool partition vector for repeated values. `dups` will have
        True for any value that is repeated anywhere else in the
        vector. It will be all False if there are no repeated values.

    Examples
    --------
    >>> from pyyeti import locate
    >>> locate.find_duplicates([0, 10, 2, 2, 6, 10, 10])
    array([False,  True,  True,  True, False,  True,  True], dtype=bool)
    """
    v = np.atleast_1d(v)
    i = np.argsort(v)
    dif = np.diff(v[i])
    dups = np.zeros(v.size, bool)
    tf = abs(dif) <= tol
    dups[i[1:-1]] = np.logical_or(tf[1:], tf[:-1])
    dups[i[0]] = tf[0]
    dups[i[-1]] = tf[-1]
    return dups


def find_subseq(seq, subseq):
    """
    Returns indices of where subseq occurs in seq.  Both are 1d numpy
    arrays.

    Parameters
    ----------
    seq : array_like
        Array to search in. It is flattened before searching.
    subseq : array_like
        Array to search for. It is flattened before searching.

    Returns
    -------
    pv : 1d ndarray
        Vector of indices:

            - length will be equal to the number of occurrences of
              `subseq`
            - the indices are to the start of each `subseq` in `seq`

        Will be empty if `subseq` is not found in `seq`.

    Examples
    --------
    >>> from pyyeti import locate
    >>> a = [1, 2, 3, 4, 5, 6, 2, 3]
    >>> sub = [2, 3]
    >>> locate.find_subseq(a, sub)  # doctest: +ELLIPSIS
    array([1, 6]...)
    >>> locate.find_subseq(a, [6, 5])
    array([], dtype=int64)
    """
    seq = np.asarray(seq).reshape(-1)
    subseq = np.asarray(subseq).reshape(-1)
    target = np.dot(subseq, subseq)
    candidates = np.where(np.correlate(seq, subseq, mode="valid") == target)[0]
    # some of the candidates entries may be false positives; check:
    check = candidates[:, np.newaxis] + np.arange(len(subseq))
    mask = np.all((np.take(seq, check) == subseq), axis=-1)
    return candidates[mask]


def find_rows(matrix, row):
    """
    Get True/False vector indicating where `row` occurs in `matrix`

    Parameters
    ----------
    matrix : array
        2d numpy array.
    row : array
        1d numpy array.

    Returns
    -------
    1d ndarray
        True/False vector with True indicating where `row` occurs in
        `matrix`. Will be all False if `row` does not occur in
        `matrix` or if ``len(row) != cols(matrix)``.

    Examples
    --------
    >>> import numpy as np
    >>> from pyyeti import locate
    >>> mat = np.array([[7, 3], [6, 8], [4, 0],
    ...                 [9, 2], [1, 5], [6, 8]])
    >>> locate.find_rows(mat,np.array([1, 2]))
    array([False, False, False, False, False, False], dtype=bool)
    >>> pv = locate.find_rows(mat,np.array([6, 8]))
    >>> pv
    array([False,  True, False, False, False,  True], dtype=bool)
    >>> mat[pv, :]
    array([[6, 8],
           [6, 8]])
    """
    c1 = np.shape(matrix)[1]
    c2 = len(row)
    if c1 != c2:
        return np.array([], dtype=int)
    # return np.nonzero(abs(matrix - row).sum(axis=1) == 0)[0]
    return abs(matrix - row).sum(axis=1) == 0


def _bytes_view(arr, dtype):
    # view columns as byte-string -- this will be sortable and make
    # the matrix look like a 1d vector; that means np.searchsorted
    # will work on it.

    # first, need to ensure c-contiguous:
    arr = np.ascontiguousarray(arr, dtype)

    # special precaution for floats:
    if np.issubdtype(arr.dtype, np.floating):
        # add zero to get rid of the minus sign on 0. ... the byte
        # string for 0. and -0. are different:

        # In [2]: a = np.array([-0.])
        # In [3]: a.view(np.dtype((np.void, 8)))
        # array([b'\x00\x00\x00\x00\x00\x00\x00\x80'], dtype='|V8')
        #
        # In [4]: a = np.array([0.])
        # In [5]: a.view(np.dtype((np.void, 8)))
        # array([b'\x00\x00\x00\x00\x00\x00\x00\x00'], dtype='|V8')

        # also:
        # In [6]: np.array([-0.]).tostring()
        # Out[6]: b'\x00\x00\x00\x00\x00\x00\x00\x80'
        # In [7]: (0. + np.array([-0.])).tostring()
        # Out[7]: b'\x00\x00\x00\x00\x00\x00\x00\x00'
        arr += 0.0
    return arr.view(np.dtype((np.void, arr.dtype.itemsize * arr.shape[-1])))


def mat_intersect(D1, D2, keep=0):
    """
    Get row intersection partition vectors between two matrices or
    vectors.

    Parameters
    ----------
    D1 : array_like
        1d or 2d array.
    D2 : array_like
        1d or 2d array.
    keep : integer
        0, 1 or 2:

          - if 0, loop over smaller matrix, finding where the rows
            occur in the larger
          - if 1, loop over `D1`, finding where the rows occur in `D2`
          - if 2, loop over `D2`, finding where the rows occur in `D1`

    Returns
    -------
    pv1 : 1d ndarray
        Row index vector into `D1`.
    pv2 : 1d ndarray
        Row index vector into `D2`.

    Notes
    -----
    `pv1` and `pv2` are found such that::

        D1[pv1] == D2[pv2]
        (Note for matrices:  M[i] == M[i, :])

    For matrices, the number of columns in `D1` and `D2` must be equal
    to get non-empty results.

    Examples
    --------
    >>> import numpy as np
    >>> from pyyeti import locate
    >>> mat1 = np.array([[7, 3], [6, 8], [4, 0], [9, 2], [1, 5]])
    >>> mat2 = np.array([[9, 2], [1, 5], [7, 3]])
    >>> pv1, pv2 = locate.mat_intersect(mat1, mat2)
    >>> pv1  # doctest: +ELLIPSIS
    array([3, 4, 0]...)
    >>> pv2  # doctest: +ELLIPSIS
    array([0, 1, 2]...)
    >>> np.all(mat1[pv1] == mat2[pv2])
    True
    >>> locate.mat_intersect(mat1, mat2, 1)  # doctest: +ELLIPSIS
    (array([0, 3, 4]...), array([2, 0, 1]...))
    >>> locate.mat_intersect(mat2, mat1, 2)  # doctest: +ELLIPSIS
    (array([2, 0, 1]...), array([0, 3, 4]...))
    >>> locate.mat_intersect(mat2, mat1)     # doctest: +ELLIPSIS
    (array([0, 1, 2]...), array([3, 4, 0]...))
    >>> mat3 = np.array([[1, 2, 3]])
    >>> locate.mat_intersect(mat1, mat3)     # doctest: +ELLIPSIS
    (array([], dtype=int...), array([], dtype=int...)
    """
    D1 = np.array(D1)
    D2 = np.array(D2)
    if D1.ndim == D2.ndim == 1:
        c1 = c2 = 1
        r1 = len(D1)
        r2 = len(D2)
        D1 = np.atleast_2d(D1)
        D2 = np.atleast_2d(D2)
        D1 = D1.T
        D2 = D2.T
    else:
        D1 = np.atleast_2d(D1)
        D2 = np.atleast_2d(D2)
        (r1, c1) = np.shape(D1)
        (r2, c2) = np.shape(D2)

    if c1 != c2:
        return np.array([], dtype=int), np.array([], dtype=int)

    # loop over the smaller one if keep == 0:
    # (we'll be searching for needles in the haystack)
    if (keep == 0 and r1 <= r2) or keep == 1:
        needles = D1
        haystack = D2
        switch = False
    else:
        needles = D2
        haystack = D1
        switch = True

    # to use the byte-string view, types must be the same:
    out_dtype = np.find_common_type([haystack.dtype, needles.dtype], [])

    # view entire rows as a single values:
    haystack = _bytes_view(haystack, out_dtype).ravel()
    needles = _bytes_view(needles, out_dtype).ravel()

    i = haystack.argsort()
    pvi = np.searchsorted(haystack, needles, sorter=i)

    # since searchsorted can return length as index:
    pvi[pvi == i.size] -= 1
    pv2 = i[pvi]

    # trim pv2 down to exact matches and create pv1:
    pv1 = np.where(haystack[pv2] == needles)[0]
    pv2 = pv2[pv1]

    if switch:
        pv1, pv2 = pv2, pv1

    return pv1, pv2


def index2bool(pv, n):
    """
    Return a True/False vector of length `n` where the True values are
    located according to `pv`.

    Examples
    --------
    >>> import numpy as np
    >>> from pyyeti import locate
    >>> pv = np.array([0, 3, 5])
    >>> locate.index2bool(pv, 8)
    array([ True, False, False,  True, False,  True, False, False], dtype=bool)
    """
    tf = np.zeros(n, dtype=bool)
    tf[pv] = True
    return tf


def flippv(pv, n):
    """Flips the meaning of an index partition vector.

    Parameters
    ----------
    pv : 1d ndarray
        The index partition to flip.
    n : integer
        The length of the dimension to partition.

    Returns
    -------
    notpv : ndarray
        Index vector; the complement of `pv`.

    Examples
    --------
    >>> import numpy as np
    >>> from pyyeti import locate
    >>> pv = np.array([0, 3, 5])
    >>> locate.flippv(pv, 8)  # doctest: +ELLIPSIS
    array([1, 2, 4, 6, 7]...)
    """
    tf = np.ones(n, dtype=bool)
    tf[pv] = False
    return tf.nonzero()[0]


def find_unique(y, tol=1e-6):
    """
    Find values in a vector that differ from previous adjacent value.

    Parameters
    ----------
    y : 1d array_like
        y-axis data vector
    tol : scalar; optional
        A value is considered the same as the previous if the
        difference is less than ``tol*max(abs(all differences))``.

    Returns
    -------
    pv : ndarray
        True/False vector with True for the unique values.

    Examples
    --------
    >>> from pyyeti import locate
    >>> locate.find_unique([1, 1, 1, 1])
    array([ True, False, False, False], dtype=bool)
    >>> locate.find_unique([4, 4, -2, -2, 0, -2])
    array([ True, False,  True, False,  True,  True], dtype=bool)
    """
    y = np.atleast_1d(y)
    m = np.diff(y)
    stol = abs(tol * abs(m).max())
    pv = np.hstack((True, abs(m) > stol))
    return pv


def list_intersect(L1, L2):
    """
    Get list intersection partition vectors between two lists

    Parameters
    ----------
    L1 : list
       List 1; the output vectors maintain the order of `L1`.
    L2 : list
       List 2.

    Returns
    -------
    pv1 : 1d ndarray
        Index vector into `L1`.
    pv2 : 1d ndarray
        Index vector into `L2`.

    Notes
    -----
    `pv1` and `pv2` are found such that::

         [L1[i] for i in pv1] == [L2[i] for i in pv2]

    Examples
    --------
    >>> from pyyeti import locate
    >>> pv1, pv2 = locate.list_intersect(['a', 3, 'z', 0],
    ...                                  [0, 'z', 1, 'a'])
    >>> pv1  # doctest: +ELLIPSIS
    array([0, 2, 3]...)
    >>> pv2  # doctest: +ELLIPSIS
    array([3, 1, 0]...)
    >>> locate.list_intersect(['a', 'b'],
    ...                       [1, 2])       # doctest: +ELLIPSIS
    (array([], dtype=int...), array([], dtype=int...))
    """
    inters = set(L1) & set(L2)
    r = len(inters)
    if r == 0:
        return np.array([], dtype=int), np.array([], dtype=int)
    pv1 = np.zeros(r, dtype=np.int64)
    pv2 = np.zeros(r, dtype=np.int64)
    j = 0
    for j, item in enumerate(inters):
        pv1[j] = L1.index(item)
        pv2[j] = L2.index(item)
    si = pv1.argsort()
    return pv1[si], pv2[si]


def merge_lists(list1, list2):
    """
    Merge two lists, trying to maintain order of both

    Parameters
    ----------
    list1 : list
        The first list to merge.
    list2 : list
        The second list to merge.

    Returns
    -------
    mlist : list
        The merged list; guaranteed to be a new list.
    pv1 : list
        List of indices specifying where the elements of `list1` are
        in `mlist`; eg: ``list1 = [mlist[i] for i in pv1]``
    pv2 : list
        List of indices specifying where the elements of `list2` are
        in `mlist`; eg: ``list2 = [mlist[i] for i in pv2]``

    Notes
    -----
    The order of `list1` is maintained. The order of `list2` is
    maintained unless it conflicts with `list1`. When a merge is
    ambiguous, the `list1` elements are merged first.

    Examples
    --------
    >>> from pyyeti import locate
    >>> l1 = ['one', 'four', 'ten']
    >>> l2 = ['zero', 'one', 'two', 'four', 'five']
    >>> l3, i, j = locate.merge_lists(l1, l2)
    >>> l3
    ['zero', 'one', 'two', 'four', 'ten', 'five']
    >>> i, j
    ([1, 3, 4], [0, 1, 2, 3, 5])
    >>> locate.merge_lists(l1, [])
    (['one', 'four', 'ten'], [0, 1, 2], [])
    """
    merged = list1[:]
    elements = []
    for e in list2:
        try:
            i = merged.index(e)
        except ValueError:
            # e is not in merged ... save it so it can be
            # inserted later, just in front of the next
            # common element
            elements.append(e)
        else:
            # e is in merged, so insert currenly accumulated
            # elements that weren't (these get inserted in
            # order, just in front of the common element e)
            for j, x in enumerate(elements):
                merged.insert(i + j, x)
            del elements[:]
    merged.extend(elements)
    # form pv1, knowing that list1 elements are in order:
    pv1 = [0] * len(list1)
    prev = 0
    for i, e in enumerate(list1):
        prev = pv1[i] = merged.index(e, prev)
    return merged, pv1, [merged.index(e) for e in list2]



# -*- coding: utf-8 -*-
import scipy.linalg as la
import numpy as np
from ._base_ode_class import _BaseODE


# FIXME: We need the str/repr formatting used in Numpy < 1.14.
try:
    np.set_printoptions(legacy="1.13")
except TypeError:
    pass


class SolveNewmark(_BaseODE):
    r"""
    2nd order ODE time domain "Newmark-Beta" solver

    This class is for solving:

    .. math::
        M \ddot{u} + B \dot{u} + K u = F

    This routine uses a fixed time step Newmark-Beta method based on
    the Nastran Theoretical Manual, section 11.4 [#newmark]_. The
    algorithm is unconditionally stable but the size of the time step
    is still a concern for accurate results. According to Bathe
    [#bathe]_, a time step that ensures accurate results for
    algorithms like this one is :math:`1/(80 f_{high})`, where
    :math:`f_{high}` is the highest frequency (Hz) in your forcing
    functions.

    In general, the mass, damping and stiffness can be fully populated
    (coupled).

    Unlike :class:`SolveUnc` and :class:`SolveExp2`, this solver can
    handle a singular mass matrix. For most (all?) other cases, the
    other two solvers are preferred since they are exact assuming
    piece-wise linear forces (if `order` is 1) or piece-wise constant
    forces (if `order` is 0). :class:`SolveUnc` is also very likely
    significantly faster for most problems.

    The equations for the non-residual equations are:

    .. math::
        A u_{n+2} = \frac{1}{3}
        \left ( F_{n+2} + F_{n+1} + F_{n} \right ) + N_{n+1} +
        A_1 u_{n+1} + A_0 u_{n}

    where:

    .. math::
        \begin{aligned}
        A &= \left [ \frac{M}{h^2} + \frac{B}{2 h} +
                     \frac{K}{3} \right ] \\
        A_1 &= \left [ \frac{2 M}{h^2} - \frac{K}{3} \right ] \\
        A_0 &= \left [ \frac{-M}{h^2} + \frac{B}{2 h} -
                       \frac{K}{3} \right ]
        \end{aligned}

    :math:`N_{n+1}` is a nonlinear force term which is optional; see
    :func:`def_nonlin`.

    To get the algorithm started, :math:`u_{-1}` and :math:`F_{-1}`
    are needed. They are computed from:

    .. math::
        \begin{aligned}
        u_{-1} &= u_0 - \dot{u}_0 h \\

        F_{-1} &= K u_{-1} + B \dot{u}_0
        \end{aligned}

    Also, :math:`F_0` is replaced by:

    .. math::
        F_{0} &= K u_{0} + B \dot{u}_0

    According to [#newmark]_, that is done to avoid ringing of
    massless degrees of freedom that are subjected to step loads. A
    side-effect of this is that :math:`\ddot{u}_0` would be zero from
    the equation of motion. However, since the acceleration is
    computed from a central finite difference formula (see below), the
    resulting initial acceleration will not be zero in general. The
    acceleration values should be approximately correct by the third
    time step.

    After the displacements have been calculated from the above
    equations, the velocities and accelerations are computed from:

    .. math::
        \begin{aligned}
        \dot{u}_n &= \frac{1}{2 h} \left (
            u_{n+1} - u_{n-1} \right ) \\

        \ddot{u}_n &= \frac{1}{h^2} \left (
            u_{n+1} - 2 u_n + u_{n-1} \right )
        \end{aligned}

    To compute the velocities and accelerations for the final time
    step, one extra integration step is performed by linearly
    extrapolating the force.

    .. note::

        The above equations are for the non-residual-flexibility
        modes. The 'rf' modes are solved statically and any initial
        conditions are ignored for them.

    References
    ----------
    .. [#newmark] 'The NASTRAN Theoretical Manual', Section 11.4,
           NASA-SP-221(06), Jan 01, 1981.
           https://ntrs.nasa.gov/search.jsp?R=19840010609

    .. [#bathe] Klaus-Jurgen Bathe, 'Finite Element Procedures',
           Second Edition, Watertown, MA: Klaus-Jürgen Bathe, 2014.
           http://web.mit.edu/kjb/www/Books/FEP_2nd_Edition_5th_Release.pdf

    See also
    --------
    :class:`SolveUnc`, :class:`SolveExp2`.

    Examples
    --------
    .. plot::
        :context: close-figs

        >>> from pyyeti import ode
        >>> import numpy as np
        >>> m = np.array([10., 30., 30., 30.])    # diag mass
        >>> k = np.array([0., 6.e5, 6.e5, 6.e5])  # diag stiffness
        >>> zeta = np.array([0., .05, 1., 2.])    # % damping
        >>> b = 2.*zeta*np.sqrt(k/m)*m            # diag damping
        >>> h = 0.0005                            # time step
        >>> t = np.arange(0, 0.2, h)              # time vector
        >>> c = 2*np.pi
        >>> f = np.vstack((3*(1-np.cos(c*2*t)),   # ffn
        ...                4.5*(1 - np.cos(np.sqrt(k[1]/m[1])*t)),
        ...                4.5*(1 - np.cos(np.sqrt(k[2]/m[2])*t)),
        ...                4.5*(1 - np.cos(np.sqrt(k[3]/m[3])*t))))
        >>> f *= 1.e4
        >>> t2 = 2 / (np.sqrt(k[1] / m[1]) / 2 / np.pi)
        >>> f[1:, t > t2] = 0.0
        >>> nb = ode.SolveNewmark(m, b, k, h)
        >>> sol = nb.tsolve(f)

        Solve with :class:`SolveUnc` for comparison:

        >>> su = ode.SolveUnc(m, b, k, h)
        >>> solu = su.tsolve(f)

        Check accuracy:

        >>> for r in 'dva':
        ...     unc = getattr(sol, r)
        ...     new = getattr(solu, r)
        ...     atol = 0.01 * abs(unc).max()
        ...     print(r + ' comp:',
        ...           np.allclose(new, unc, atol=atol, rtol=0.001))
        d comp: True
        v comp: True
        a comp: True

        Plot the four accelerations:

        >>> import matplotlib.pyplot as plt
        >>> fig = plt.figure('Example', figsize=[8, 8])
        >>> fig.clf()
        >>> labels = ['Rigid-body', 'Underdamped',
        ...           'Critically Damped', 'Overdamped']
        >>> for j, name in zip(range(4), labels):
        ...     _ = plt.subplot(4, 1, j+1)
        ...     _ = plt.plot(t, solu.a[j], label='SolveUnc')
        ...     _ = plt.plot(t, sol.a[j], '--', label='SolveNewmark')
        ...     _ = plt.title(name)
        ...     _ = plt.ylabel('Acceleration')
        ...     _ = plt.xlabel('Time (s)')
        ...     if j == 0:
        ...         _ = plt.legend(loc='best')
        >>> fig.tight_layout()
    """

    def __init__(self, m, b, k, h=None, rf=None):
        r"""
        Instantiates a :class:`SolveNewmark` solver.

        Parameters
        ----------
        m : 1d or 2d ndarray or None
            Mass; vector (of diagonal), or full; if None, mass is
            assumed identity
        b : 1d or 2d ndarray
            Damping; vector (of diagonal), or full
        k : 1d or 2d ndarray
            Stiffness; vector (of diagonal), or full
        h : scalar or None; optional
            Time step; can be None if only want to solve a static
            problem or if only solving frequency domain problems
        rf : 1d array or None; optional
            Index or bool partition vector for residual-flexibility
            modes; these will be solved statically. The `rf` option
            only applies to modal space equations.

        Notes
        -----
        The instance is populated with some or all of the following
        members.

        ============  ================================================
        Member        Description
        ============  ================================================
        m             mass for the non-rf DOF (or None for identity)
        b             damping for the non-rf DOF
        k             stiffness for the non-rf DOF
        h             time step
        rb            np.array([])
        el            index vector or slice for the non-rf DOF
        rf            index vector or slice for the rf DOF
        nonrf         index vector or slice for the non-rf DOF
        kdof          index vector or slice for the non-rf DOF
        n             number of total DOF
        rbsize        0
        elsize        number of non-rf DOF
        rfsize        number of rf DOF
        nonrfsz       number of non-rf DOF
        ksize         number of non-rf DOF
        krf           stiffness for the rf DOF
        ikrf          inverse of stiffness for the rf DOF
        unc           True if there are no off-diagonal terms in any
                      matrix; False otherwise
        systype       float or complex; determined by `m` `b` `k`
        Ad            decomposed version of matrix :math:`A` (see
                      :class:`SolveNewmark`)
        A0            decomposed version of matrix :math:`A_0`
        A1            decomposed version of matrix :math:`A_1`
        nonlin_terms  number of nonlinear force terms defined by
                      :func:`def_nonlin` (initially set to 0)
        ============  ================================================
        """
        self._common_precalcs(m, b, k, h, rb=[], rf=rf)
        self._newmark_precalcs()
        self._mk_slices()  # dorbel=False)

    def tsolve(self, force, d0=None, v0=None):
        """
        Solve time-domain 2nd order ODE equations

        Parameters
        ----------
        force : 2d ndarray
            The force matrix; ndof x time
        d0 : 1d ndarray; optional
            Displacement initial conditions; if None, zero ic's are
            used.
        v0 : 1d ndarray; optional
            Velocity initial conditions; if None, zero ic's are used.

        Returns
        -------
        A record (SimpleNamespace class) with the members:

        d : 2d ndarray
            Displacement; ndof x time
        v : 2d ndarray
            Velocity; ndof x time
        a : 2d ndarray
            Acceleration; ndof x time
        h : scalar
            Time-step
        t : 1d ndarray
            Time vector: np.arange(d.shape[1])*h
        z : dictionary of 2d ndarrays; optional
            Only present if there are nonlinear force terms. The
            dictionary keys correspond to the keys used to define the
            nonlinear force terms in :func:`def_nonlin`. The values
            are the output of the functions that defined these force
            terms (see :func:`def_nonlin`).
        """
        force = np.atleast_2d(force)
        d, v, a, F = self._init_dva(force, d0, v0)
        if self.ksize:
            D = d[self.kdof]
            V = v[self.kdof]
            A = a[self.kdof]
            nt = D.shape[1]
            h = self.h
            h2 = 2 * h
            sqh = h * h
            A1 = self.A1
            A0 = self.A0
            # For the last velocity and acceleration, extrapolate
            # force to run 1 more step (so we can calculate the final
            # v and a): Fe = 2*F[:, -1] - F[:, -2] ... when added to
            # other 2 force terms (F[:, -1] + F[:, -2]), we get 3 *
            # F[:, -1].
            if self.nonlin_terms == 0:
                if self.unc:
                    for j in range(2, nt):
                        D[:, j] = (
                            F[:, j]
                            + F[:, j - 1]
                            + F[:, j - 2]
                            + A1 * D[:, j - 1]
                            + A0 * D[:, j - 2]
                        )
                    De = 3 * F[:, -1] + A1 * D[:, -1] + A0 * D[:, -2]
                else:
                    for j in range(2, nt):
                        D[:, j] = (
                            F[:, j]
                            + F[:, j - 1]
                            + F[:, j - 2]
                            + A1 @ D[:, j - 1]
                            + A0 @ D[:, j - 2]
                        )
                    De = 3 * F[:, -1] + A1 @ D[:, -1] + A0 @ D[:, -2]
            else:

                def _get_nonlin(j):
                    N = 0.0
                    for key, (func, T, args) in self.nl_dct.items():
                        z = func(D, j, h, **args)
                        self.z[key][:, j] = z
                        N += T @ z
                    return N

                if self.unc:
                    for j in range(2, nt):
                        D[:, j] = (
                            F[:, j]
                            + F[:, j - 1]
                            + F[:, j - 2]
                            + _get_nonlin(j - 1)
                            + A1 * D[:, j - 1]
                            + A0 * D[:, j - 2]
                        )
                    De = (
                        3 * F[:, -1]
                        + _get_nonlin(nt - 1)
                        + A1 * D[:, -1]
                        + A0 * D[:, -2]
                    )
                else:
                    for j in range(2, nt):
                        D[:, j] = (
                            F[:, j]
                            + F[:, j - 1]
                            + F[:, j - 2]
                            + _get_nonlin(j - 1)
                            + A1 @ D[:, j - 1]
                            + A0 @ D[:, j - 2]
                        )
                    De = (
                        3 * F[:, -1]
                        + _get_nonlin(nt - 1)
                        + A1 @ D[:, -1]
                        + A0 @ D[:, -2]
                    )

            # calculate velocity and acceleration
            V[:, 1:-1] = (D[:, 2:] - D[:, :-2]) / h2
            V[:, -1] = (De - D[:, -2]) / h2
            A[:, 1:-1] = (D[:, 2:] - 2 * D[:, 1:-1] + D[:, :-2]) / sqh
            A[:, -1] = (De - 2 * D[:, -1] + D[:, -2]) / sqh

            if not self.slices:
                d[self.kdof] = D
                v[self.kdof] = V
                a[self.kdof] = A

        sol = self._solution(d, v, a)
        if self.nonlin_terms:
            sol.z = self.z
        return sol

    def def_nonlin(self, dct):
        r"""
        Define nonlinear force terms

        Parameters
        ----------
        dct : dictionary

            Dictionary where each entry defines a nonlinear force
            term. You can define any number of terms. Each value is a
            sequence of 2 or 3 items: a function (or other callable),
            a 2d ndarray, and optionally a dictionary of arguments for
            the function. The dictionary keys are arbitrary but will
            be used as keys in the `z` return value in the output of
            :func:`tsolve`. The form of the dictionary is::

               {
                   key_0: (func_0, T_0 [, optargs_0]),
                   key_1: (func_1, T_1 [, optargs_1]),
                   ...
               }

            =========  ===============================================
            Item       Description
            =========  ===============================================
            func_i     Each function must accept at least 3 arguments:
                       the current solution displacement matrix (`d`),
                       the current step index (`j`), and the time step
                       (`h`). It can accept other arguments which are
                       specified via `optargs`::

                           def func_i(d, j, h [, ...]):

                       The function must return a 1d ndarray. It will
                       be multiplied by the corresponding transform
                       `T` and added to the other force terms (this is
                       the :math:`N_{n+1}` term shown in
                       :class:`SolveNewmark`). See note below if
                       velocity dependent terms are needed.

            T_i        Each `T_i` matrix transforms the corresponding
                       function output to a force. If your function
                       outputs the final force already, just use
                       identity for this transform. The reason this
                       input is here, instead of relying on the
                       user-supplied function `func_i` to apply the
                       transform internally, is for efficiency: the
                       relatively expensive operation of:
                       :math:`A^{-1} T_i` is done now, outside the
                       integration loop. The matrix :math:`A` is
                       defined in :class:`SolveNewmark`.

            optargs_i  If included, each `optargs_i` is a dictionary
                       of arbitrary arguments for `func_i`.
            =========  ===============================================

        Notes
        -----
        The the `j`'th nonlinear force term is computed by::

            N_j = (T_0 @ func_0(d, j, h, **optargs_0) +
                   T_1 @ func_1(d, j, h, **optargs_1) +
                   ...)

        That term is used to compute the ``j+1``'th displacement; see
        equation in :class:`SolveNewmark`. That is appropriate because
        of the nature of the central finite difference formulae used
        in the Newmark-Beta formation.

        The solution namespace returned by :func:`tsolve` will contain
        the outputs of each user defined function in the dictionary
        `z`.

        .. note::

            Nastran estimates velocity by: :math:`v_j = (d_j -
            d_{j-1})/h`. In your function, you can use something like:
            ``vj = (d[:, j] - d[:, j-1])/h``. That will work even for
            the first call, when `j` is 0, because "-1" displacements
            are stored in the last column of `d` for just this
            purpose.

        .. note::

            The nonlinear forces are computed in the integration loop
            for the non-residual flexibility equations only.
            Therefore, it is recommended to not use the `rf` option
            with nonlinear force terms.

        Examples
        --------
        Model a two-mass system with one linear spring and one
        nonlinear spring. The nonlinear spring is only active when
        compressed. There is a gap of 0.01 units before the spring
        starts being compressed.

        Model::

              |--> x1        |--> x2
            |----|    50   |----|
            | 10 |--\/\/\--| 12 |   F(t)
            |    |         |    | =====>
            |----| |-/\/-| |----|
                 K_nonlinear

            F(t) = 5000 * np.cos(2 * np.pi * t + 270 / 180 * np.pi)

        The nonlinear spring force is linearly interpolated according
        to the "lookup" table below. Linear extrapolation is used for
        displacements out of range of the table.

        .. plot::
            :context: close-figs

            >>> import numpy as np
            >>> from scipy.interpolate import interp1d
            >>> import matplotlib.pyplot as plt
            >>> from pyyeti import ode
            >>>
            >>> # mass and stiffness:
            >>> m = np.diag([10., 12.])
            >>> k = np.array([[50., -50.],
            ...               [-50.,  50.]])
            >>> c = 0. * k  # no damping
            >>>
            >>> # define time steps and force:
            >>> h = 0.005
            >>> t = np.arange(0, 4 + h / 2, h)
            >>> f = np.zeros((2, t.size))
            >>> f[1] = 5000 * np.cos(2 * np.pi * t + 3 / 2 * np.pi)
            >>>
            >>> # define interpolation table for force (the lookup
            >>> # value is x1 - x2):
            >>> lookup = np.array([[-10,     0.],
            ...                    [0.01,    0.],
            ...                    [5,     200.],
            ...                    [6,    1000.],
            ...                    [10,   1500.]])
            >>>
            >>> # force transforming lookup value to forces on the
            >>> # masses:
            >>> Tfrc = np.array([[-1.], [1.]])
            >>>
            >>> # turn interpolation table into a function for speed
            >>> interp_func = interp1d(*lookup.T,
            ...                        fill_value='extrapolate')
            >>>
            >>> # function needed for ode.SolveNewmark.def_nonlin:
            >>> def nonlin(d, j, h, interp_func):
            ...     return interp_func(d[[0], j] - d[[1], j])
            >>>
            >>> # Solve:
            >>> ts = ode.SolveNewmark(m, c, k, h)
            >>> ts.def_nonlin(
            ...     {'kcomp': (nonlin, Tfrc,
            ...               dict(interp_func=interp_func))})
            >>> sol = ts.tsolve(f)
            >>>
            >>> # for comparison, run in SolveExp2 using the generator
            >>> # feature:
            >>> ts2 = ode.SolveExp2(m, c, k, h)
            >>> gen, d, v = ts2.generator(len(t), f[:, 0])
            >>>
            >>> for i in range(1, len(t)):
            ...     if i == 1:
            ...         dx = d[0, i - 1] - d[1, i - 1]
            ...     else:
            ...         # for improved convergence, use linear
            ...         # interpolation to estimate displacements at
            ...         # current time:
            ...         dx = (2 * d[0, i - 1] - d[0, i - 2] +
            ...               d[1, i - 2] - 2 * d[1, i - 1])
            ...     f_nl = (Tfrc @ interp_func([dx]))
            ...     gen.send((i, f[:, i] + f_nl))
            >>>
            >>> sol2 = ts2.finalize()
            >>>
            >>> # plot results:
            >>> _ = plt.figure('Example', figsize=(8, 8))
            >>> plt.clf()
            >>>
            >>> _ = plt.subplot(311)
            >>> _ = plt.plot(t, sol.d.T)
            >>> _ = plt.plot(t, d.T, '--')
            >>> _ = plt.title('x1 and x2 displacments')
            >>> _ = plt.ylabel('Displacement')
            >>> _ = plt.legend(('SolveNewmark x1',
            ...                 'SolveNewmark x2',
            ...                 'SolveExp2 x1',
            ...                 'SolveExp2 x2'))
            >>>
            >>> _ = plt.subplot(312)
            >>> _ = plt.plot(t, sol.d[0] - sol.d[1],
            ...              label='SolveNewmark')
            >>> _ = plt.plot(t, d[0] - d[1], '--', label='SolveExp2')
            >>> _ = plt.title('Relative displacement: x1 - x2')
            >>> _ = plt.ylabel('Displacement')
            >>> _ = plt.legend()
            >>>
            >>> _ = plt.subplot(313)
            >>> _ = plt.plot(t, sol.z['kcomp'][0],
            ...              label='SolveNewmark')
            >>> _ = plt.plot(t, interp_func(d[0, :] - d[1, :]), '--',
            ...              label='SolveExp2')
            >>> _ = plt.title('Force in nonlinear spring')
            >>> _ = plt.xlabel('Time (s)')
            >>> _ = plt.ylabel('Force')
            >>> _ = plt.legend()
            >>> _ = plt.tight_layout()
        """
        # apply inv(A) to the transforms while making new dict:
        nl_dct = {}
        for k, v in dct.items():
            if len(v) == 2:
                args = {}
            else:
                args = v[2]

            if self.unc:
                T = v[1] / self.Ad[:, None]
            else:
                T = la.lu_solve(self.Ad, v[1])

            nl_dct[k] = (v[0], T, args)

        # if here, everything must be okay:
        self.nonlin_terms = len(nl_dct)
        self.nl_dct = nl_dct

    def _newmark_precalcs(self):
        # setup matrices for newmark - beta solution beta = 1/3
        # A u_(n + 2) = 1 / 3 * (F_(n + 2) + F_(n + 1) + F_(n)) +
        #               NonLin_(n + 1) + A1 u_(n + 1) + A0 u_(n)
        #
        # where:
        #  A = 1 / h ^ 2 M + 1 / (2 h) B + 1 / 3 K
        #  A1 = 2 / h ^ 2 M - 1 / 3 K
        #  A0 = -1 / h ^ 2 M + 1 / (2 h) B - 1 / 3 K
        self.pc = True  # to make _alloc_dva happy
        self.nonlin_terms = 0
        if self.ksize == 0:
            return
        h = self.h
        sqh = h * h
        h2 = 2 * h

        if self.m is None:
            if self.unc:
                # have diagonals:
                mterm = 1.0 / sqh
            else:
                # have matrices:
                mterm = np.diag(np.ones(self.ksize) / sqh)
        else:
            mterm = self.m / sqh

        A = mterm + self.b / h2 + self.k / 3
        A1 = 2 * mterm - self.k / 3
        A0 = self.b / h2 - self.k / 3 - mterm

        if self.unc:
            # have diagonals:
            self.Ad = A

            # define new A1, A0:
            self.A0 = A0 / A
            self.A1 = A1 / A
        else:
            # have square matrices:
            self.Ad = la.lu_factor(A, overwrite_a=True)

            # define new A1, A0:
            self.A0 = la.lu_solve(self.Ad, A0, overwrite_b=True)
            self.A1 = la.lu_solve(self.Ad, A1, overwrite_b=True)

    def _init_dva(self, force, d0, v0):
        """
        Newmark Beta version of _init_dva
        """
        if force.shape[0] != self.n:
            raise ValueError(
                f"Force matrix has {force.shape[0]} rows; {self.n} rows are expected"
            )

        d0, v0 = self._set_initial_cond(d0, v0)
        nt = force.shape[1]
        d, v, a = self._alloc_dva(nt, True)

        # solve resflex part:
        if self.rfsize:
            if self.unc:
                d[self.rf] = self.ikrf * force[self.rf]
            else:
                d[self.rf] = la.lu_solve(self.ikrf, force[self.rf])
            force = force[self.nonrf]

        if self.ksize == 0:
            return d, v, a, force

        self._init_dv(d, v, d0, v0, F0=None, static_ic=False)
        d0 = np.zeros(self.ksize) if d0 is None else d0[self.nonrf]
        v0 = np.zeros(self.ksize) if v0 is None else v0[self.nonrf]

        # to get the algorithm going and stable (see Nastran
        # theoretical manual, section 11.4):
        force = force / 3.0
        h = self.h
        u_1 = d0 - v0 * h

        # compute nonlinear force terms:
        N = 0.0
        if self.nonlin_terms:
            d[self.nonrf, -1] = u_1
            self.z = {}
            for key, (func, T, args) in self.nl_dct.items():
                z0 = func(d, 0, h, **args)
                z = np.empty((z0.shape[0], nt))
                z[:, 0] = z0
                self.z[key] = z
                N += T @ z0

        if self.unc:
            force[:, 0] = (self.k * d0 + self.b * v0) / 3.0
            # method 2 from theory manual:
            # force[:, 0] = (3 * force[:, 0] + self.k * d0 +
            #                self.b * v0) / 6.0
            force /= self.Ad[:, None]
            F_1 = (self.k * u_1 + self.b * v0) / (3 * self.Ad)

            # because of the - 1 subscript, do first step outside of
            # the loop:
            d[self.nonrf, 1] = (
                force[:, 1] + force[:, 0] + F_1 + N + self.A1 * d0 + self.A0 * u_1
            )
        else:
            force[:, 0] = (self.k @ d0 + self.b @ v0) / 3.0
            # force[:, 0] = (3 * force[:, 0] + self.k @ d0 +
            #                self.b @ v0) / 6.0
            force = la.lu_solve(self.Ad, force, overwrite_b=True)
            F_1 = la.lu_solve(
                self.Ad, (self.k @ u_1 + self.b @ v0) / 3, overwrite_b=True
            )

            # because of the - 1 subscript, do first step outside of
            # the loop:
            d[self.nonrf, 1] = (
                force[:, 1] + force[:, 0] + F_1 + N + self.A1 @ d0 + self.A0 @ u_1
            )

        a[self.nonrf, 0] = (d[self.nonrf, 1] - 2 * d0 + u_1) / (h * h)
        return d, v, a, force



import matplotlib.pyplot as plt
from matplotlib.widgets import Slider, Button


class ColumnPlotter:
    r"""
    A very basic class to visually compare matrices by plotting
    columns in an interactive window.

    The user can scroll forward and backward through the columns of
    data, while using the plotting tools provided by `matplotlib
    <https://matplotlib.org>`_ (such as zoom).

    Examples
    --------
    .. plot::
        :context: close-figs

        >>> import numpy as np
        >>> from pyyeti.column_plotter import ColumnPlotter
        >>> #
        >>> h = 0.01
        >>> t = np.arange(0, 1, h)
        >>> sin = np.sin(3 * 2 * np.pi * t)
        >>> sin_a = sin + np.random.randn(*sin.shape) * 0.1
        >>> sin_b = sin + np.random.randn(*sin.shape) * 0.1
        >>> A = np.column_stack((sin, sin_a, sin_b))
        >>> B = A + np.random.randn(*A.shape) * 0.2 + 2
        >>> C = A + np.random.randn(*A.shape) * 0.2 + 3
        >>> #
        >>> cp = ColumnPlotter(
        ...     t,
        ...     dict(A=A, B=B, C=C),
        ...     ["1st column", "2nd column", "3rd column"],
        ... )
    """

    def __init__(self, x, dct, column_labels=None):
        """
        Instantiates a :class:`ColumnPlotter` object

        Parameters
        ----------
        x : 1d array_like
            The x-axis data
        dct : dict
            Dictionary of y-axis data. The key is the label for the
            data and will be in the legend. The values are the
            matrices to compare and the number of rows must equal
            ``len(x)``. All matrices are expected to be the same size.
        column_labels : list or None; optional
            List of strings to be used for the plot titles;
            ``len(column_labels)`` is expected to be equal to the
            number of columns in the y-axis data matrices.
        """
        self.x = x
        self.dct = dct
        self.column_labels = column_labels
        self.ind = 0
        fig, self.ax = plt.subplots()
        plt.subplots_adjust(bottom=0.2)

        self.lines = {}
        self.n_columns = None
        for key, value in self.dct.items():
            line = plt.plot(self.x, value[:, self.ind], label=key)[0]
            self.lines[line] = value
            if self.n_columns is None:
                self.n_columns = value.shape[1]

        plt.legend()
        plt.grid(True)
        if self.column_labels:
            self.ttl = plt.title(self.column_labels[self.ind])

        axprev = plt.axes([0.7, 0.05, 0.1, 0.075])
        axnext = plt.axes([0.81, 0.05, 0.1, 0.075])
        self.bnext = Button(axnext, "Next")
        self.bnext.on_clicked(self._next)
        self.bprev = Button(axprev, "Previous")
        self.bprev.on_clicked(self._prev)

        axcolor = "lightgoldenrodyellow"
        slider_ax = plt.axes([0.1, 0.1, 0.50, 0.03], facecolor=axcolor)
        self.slider = Slider(
            slider_ax, "Column", 0, self.n_columns - 1, valinit=0, valfmt="%.0f"
        )
        self.slider.on_changed(self._update_column)

    def _update_plot(self):
        for line, value in self.lines.items():
            line.set_ydata(value[:, self.ind])
        self.ax.relim()
        # update ax.viewLim using the new dataLim
        self.ax.autoscale()
        # self.ax.autoscale_view()
        if self.column_labels:
            self.ttl.set_text(self.column_labels[self.ind])
        plt.draw()

    def _next(self, event):
        # print("next")
        if self.ind < self.n_columns - 1:
            self.ind += 1
            self.slider.set_val(self.ind)
            self._update_plot()

    def _prev(self, event):
        # print("prev")
        if self.ind > 0:
            self.ind -= 1
            self.slider.set_val(self.ind)
            self._update_plot()

    def _update_column(self, new_column):
        self.ind = int(new_column)
        self._update_plot()
